/*
 * Demo Android application for Tellu Smart Platform REST API
 * library.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
package no.tellu.findit.client.android.demo;

import java.util.Map;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import no.tellu.findit.client.api.AsyncService;
import no.tellu.findit.client.api.DeleteTransaction;
import no.tellu.findit.client.api.GetTransaction;
import no.tellu.findit.client.api.PostTransaction;
import no.tellu.findit.client.api.PutTransaction;
import no.tellu.findit.client.api.ResourceTransaction;
import no.tellu.findit.client.api.Resources;
import no.tellu.findit.client.edge.AsyncEdgePoster;
import no.tellu.findit.client.edge.Observation;
import no.tellu.findit.client.edge.ObservationPoster;
import android.util.Log;
import android.widget.TextView;

public class Transactions {
	public static final String LOG_TAG = "SmartTracker";
	
	private static final String ASSET_NAME = "APITest8261";
	
	final AsyncService apiService;
	final TextView log;
	
	ResourceTransaction currentTrans;
	AsyncService.TransactionCallback transCallback;
	AsyncService.DataRetrieveCallback retrieveCallback;
	
	/*
	 * Note on IDs: String resourceId is the preferred way to target
	 * resource objects, rather than the numerical, internal id,
	 * but there is a bug with DELETE using resourceId in the current
	 * version of the REST API, hence the use of numerical id here.
	 */
	long assetId = -1;
	long newAssetId;
	boolean assetFound;
	long deviceId;
	
	ObservationPoster edgePoster;
	
	
	public Transactions(AsyncService service, TextView v) {
		apiService = service;
		log = v;
	}
	
	private void logAction(String message) {
		log.append(message + "\n");
		Log.i(LOG_TAG, message);
	}
	
	private void logResult(String message) {
		log.append("." + message + "\n");
		Log.d(LOG_TAG, message);
	}
	
	private void logError(String message, Exception e) {
		log.append("." + message + "\n");
		if (e == null)
			Log.e(LOG_TAG, message);
		else
			Log.e(LOG_TAG, message, e);
	}
	
	private void logError(int code, String message, Exception e) {
		log.append(".Transaction failed with code " + code + "\n");
		log.append(".Error: " + message + "\n");
		if (e == null)
			Log.e(LOG_TAG, "Transaction error: " + message);
		else
			Log.e(LOG_TAG, "Transaction error: " + message, e);
	}
	
	/**
	 * Runs through all transactions specified in this class.
	 */
	public void doDemoTransactions() {
		postAsset();
	}
	
	/**
	 * Create new Asset.
	 */
	void postAsset() {
		logAction("POST Asset");
		transCallback = new AsyncService.TransactionCallback() {
			@Override
			public void onTransactionOK(JSONObject object) {
				try {
					newAssetId = object.getLong("id");
					logResult("Asset created with id " + newAssetId);
					putAsset();
				} catch (JSONException e) {
					logError("JSON error", e);
				}
			}
			
			@Override
			public void onTransactionError(int code, String message, Exception e) {
				logError(code, message, e);
			}
		};
		PostTransaction post = new PostTransaction(Resources.ASSET);
		//Can specify object using JSONObject or String
		post.setObject("{\"name\":\"Test4kgf73m\"}");
		currentTrans = post;
		apiService.doTransaction(currentTrans, transCallback);
	}
	
	/**
	 * Change name of new Asset.
	 */
	void putAsset() {
		logAction("PUT Asset name");
		transCallback = new AsyncService.TransactionCallback() {
			@Override
			public void onTransactionOK(JSONObject object) {
				try {
					String name = object.getString("name");
					logResult("Name changed to " + name);
					getAssets();
				} catch (JSONException e) {
					logError("JSON error", e);
				}
			}
			
			@Override
			public void onTransactionError(int code, String message, Exception e) {
				logError(code, message, e);
			}
		};
		try {
			JSONObject asset = new JSONObject();
			asset.put("id", newAssetId);
			asset.put("name", ASSET_NAME);
			PutTransaction put = new PutTransaction(Resources.ASSET);
			put.setObject(asset);
			currentTrans = put;
			apiService.doTransaction(currentTrans, transCallback);
			
		} catch (JSONException e) {
			logError("JSON error", e);
		}
	}

	/**
	 * Get list of all Assets.
	 */
	void getAssets() {
		logAction("GET Assets (*: has device)");
		assetFound = false;
		assetId = -1;
		deviceId = -1;
		retrieveCallback = new AsyncService.DataRetrieveCallback() {
			@Override
			public void onRetrieveOK(JSONArray objects, GetTransaction transaction) {
				JSONObject jo;
				String name;
				for (int i=0; i<objects.length(); i++) {
					try {
						jo = objects.getJSONObject(i);
						name = jo.getString("name");
						if (name.equals(ASSET_NAME))
							assetFound = true;
						else if (assetId == -1)
							assetId = jo.getLong("id");
						boolean hasDevice = !(jo.isNull("positionProvider"));
						if (hasDevice) {
							logResult(name + "*");
							if (deviceId == -1) {
								JSONObject pp = jo.getJSONObject("positionProvider");
								deviceId = pp.getLong("id");
							}
						} else {
							logResult(name);
						}
					} catch (JSONException e) {
						logError("JSON error", e);
					}
				}
				if (!doGet(transaction)) {
					//Finished
					if (assetFound)
						logResult("New asset found");
					else
						logError("New Asset NOT found!", null);
					deleteAsset();
				}
			}
			
			@Override
			public void onRetrieveError(int code, String message, GetTransaction transaction) {
				logError(code, message, transaction.getException());
			}
		};
		GetTransaction trans = new GetTransaction(Resources.ASSET);
		trans.setContent("positionProvider"); //Include positionProvider field
		currentTrans = trans;
		doGet(trans);
	}
	
	boolean doGet(GetTransaction transaction) {
		if (transaction.next()) {
			if (transaction.getTotal() > -1)
				logAction("GET more items, total is " + transaction.getTotal());
			apiService.doDataRetrieve(transaction, retrieveCallback);
			//Either retrieveCallback.onRetrieveOK or retrieveCallback.onRetrieveError will be called
			return true;
			
		} else {
			currentTrans = null;
			retrieveCallback = null;
			return false;
		}
	}
	
	/**
	 * Delete the new Asset.
	 */
	void deleteAsset() {
		logAction("DELETE Asset " + newAssetId);
		transCallback = new AsyncService.TransactionCallback() {
			@Override
			public void onTransactionOK(JSONObject object) {
				logResult("Asset deleted OK");
				if (assetId == -1)
					logAction("No assets, DONE");
				else
					getSingleAsset();
			}
			
			@Override
			public void onTransactionError(int code, String message, Exception e) {
				logError(code, message, e);
			}
		};
		DeleteTransaction delete = new DeleteTransaction(Resources.ASSET, newAssetId);
		currentTrans = delete;
		apiService.doTransaction(currentTrans, transCallback);
	}
	
	/**
	 * Get all data on a single Asset.
	 */
	void getSingleAsset() {
		logAction("GET Asset " + assetId);
		transCallback = new AsyncService.TransactionCallback() {
			@Override
			public void onTransactionOK(JSONObject object) {
				try {
					logResult("Got Asset " + object.getString("name"));
					if (!object.isNull("type")) {
						JSONObject jo = object.getJSONObject("type");
						logResult("Asset type " + jo.getString("name"));
					}
					if (object.isNull("lastPosition"))
						logResult("Asset has no position");
					else
						logResult("Asset has position");
					JSONArray ja = object.getJSONArray("properties");
					if (ja == null) {
						logResult("Asset has no properties");
					} else {
						logResult("Number of properties: " + ja.length());
					}
					if (deviceId > -1)
						getDevice();
					else
						logAction("DONE");
				} catch (JSONException e) {
					logError("JSON error", e);
				}
			}
			
			@Override
			public void onTransactionError(int code, String message, Exception e) {
				logError(code, message, e);
			}
		};
		GetTransaction get = new GetTransaction(Resources.ASSET);
		get.setId(assetId);
		currentTrans = get;
		apiService.doTransaction(currentTrans, transCallback);
	}
	
	void getDevice() {
		logAction("GET Device " + deviceId);
		transCallback = new AsyncService.TransactionCallback() {
			@Override
			public void onTransactionOK(JSONObject object) {
				try {
					logResult("Got Device " + object.getString("name"));
					String uuid = object.getString("uuid");
					postObservation(uuid);
					
				} catch (JSONException e) {
					logError("JSON error", e);
				}
			}
			
			@Override
			public void onTransactionError(int code, String message, Exception e) {
				logError(code, message, e);
			}
		};
		GetTransaction get = new GetTransaction(Resources.DEVICE);
		get.setId(deviceId);
		currentTrans = get;
		apiService.doTransaction(currentTrans, transCallback);
	}
	
	void postObservation(String uuid) {
		logAction("POST Observation to Device");
		edgePoster = new AsyncEdgePoster(DemoActivity.EDGE_URL, uuid);
		
		ObservationPoster.PostCallback postCallback = new ObservationPoster.PostCallback() {
			@Override
			public void onPostOK(int code, Map<String, String> response) {
				logResult("Observation POST OK");
				logAction("DONE");
			}
			
			@Override
			public void onPostError(int code, Exception e) {
				logError(code, "Edge POST error", e);
			}
		};
		
		Observation obs = new Observation();
		obs.setPropertyAsString("testProp1", "1");
		obs.setPropertyAsString("testProp2", "something");
		edgePoster.doPostAsync(obs, postCallback);
	}
}
